import sys
import argparse
from torch.distributed.rpc import RRef, rpc_async, remote
import torch.distributed.rpc as rpc
import torch
from part.Utils import GraphData
from Sample.Sampler import NeighborSampler 
import time
from typing import Optional
import torch.distributed as dist

from BatchData import BatchData

OBSERVER_NAME='server{}'
sampler_workers={}
distributed_loader_list={}
local_sample_worker={}

class sampler_worker:
    def __init__(self,worker_size,OBSERVER_NAME):
        self.id = rpc.get_worker_info().id
        self.rank=self.id
        self.worker_size=worker_size
        sampler_workers[self.rank]=self

        #for ob_rank in range(worker_size):
        #    if(ob_rank == self.rank):
        #        self.ob_rrefs.append(RRef(self))
        #        continue
        #    ob_info = rpc.get_worker_info(OBSERVER_NAME.format(ob_rank))
        #    self.ob_rrefs.append(remote(ob_info, sampler_worker,args=(worker_size,OBSERVER_NAME)))
        #    futs=[]
        if(self.rank==0):
            futs=[]
            self.ob_rrefs=[RRef(self)]
            for rank in range(1,worker_size):
                ob_info = rpc.get_worker_info(OBSERVER_NAME.format(rank))
                self.ob_rrefs.append(remote(ob_info, sampler_worker,args=(worker_size,OBSERVER_NAME)))
            for rank in range(1,worker_size):
                ob_rref=self.ob_rrefs[rank]
                futs.append(
                    rpc_async(
                        ob_rref.owner(),
                        ob_rref.rpc_sync().init_rref,
                        args=(self.ob_rrefs,)
                    )
                )
            for fut in futs:
                fut.wait()


    def init_rref(self,rrefs):
        self.ob_rrefs=rrefs
        local_sample_worker[self.rank]=self
    
    def request_attr(self,data_name,node_feature):
        print('request',self.rank,node_feature)
        futs=[]
        feature_part=[]
        ob_rrefs=self.ob_rrefs
        for i in range(self.worker_size):
            if(i==self.rank):
                continue
            futs.append(
                rpc_async(
                        ob_rrefs[i].owner(),
                        ob_rrefs[i].rpc_sync().get_localattr,
                        args=(data_name,node_feature[i],)
                    )
            )
        
        for fut in futs:
            feature_part.append(fut.wait())
        print('request end')
        return feature_part
    
    def get_localattr(self,data_name,node_list):
        print('local',self.rank,node_list)
        for key in distributed_loader_list:
            print('key',key)
        graph_loader=distributed_loader_list[(self.rank,data_name)]
        print('local end')
        return graph_loader.get_localattr(node_list)

parser = argparse.ArgumentParser(
    description="RPC Reinforcement Learning Example",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
)

parser.add_argument('--world_size', default=4, type=int, metavar='W',
                    help='number of workers')
parser.add_argument('--log_interval', type=int, default=10, metavar='N',
                    help='interval between training status logs')
parser.add_argument('--gamma', type=float, default=0.99, metavar='G',
                    help='how much to value future rewards')
parser.add_argument('--seed', type=int, default=1, metavar='S',
                    help='random seed  for reproducibility')
args = parser.parse_args()

def init_sampler_worker():
    
    local_sample_worker[0]=sampler_worker(args.world_size,OBSERVER_NAME)


class DistributedDataLoader:
    ''' 
     Args:
            data_path: the path of loaded graph ,each part 0 of graph is saved on $path$/rank_0
            num_replicas: the num of worker
            

    '''
    def __init__(self,data_name,data_path,sampler,num_samples: Optional[int]=None,shuffle:bool = True,seed:int=0,**kwargs):
        
        self.data_name=data_name
        self.num_samples=num_samples
        self.sampler=sampler
        self.epoch=0
        self.id = rpc.get_worker_info().id
        self.rank=self.id
        self.worker_size=args.world_size#sampler_workers[self.rank].worker_size
        print('load graph',data_path+'/rank_{}'.format(self.rank))
        self.graph=GraphData(data_path+'/rank_{}'.format(self.rank))
        self.shuffle=shuffle
        self.seed=seed
        self.kwargs=kwargs
        distributed_loader_list[(self.rank,self.data_name)]=self
        
        dist.barrier()
        
    def __next__(self):
        print(self.sampler,self.num_samples,self.kwargs)
        #self.sampleGraph(self.sampler,self.num_samplers,self.kwargs)
        return self.sampleGraph(self.sampler,self.num_samples,**self.kwargs)
        
    def set_epoch(self,epoch):
        self.epoch = epoch 

    #将所需node_list的节点拆成各分区的节点
    def split_node_part(self,node_id_list):
        part_list=[]
        scatter_list=[]
        for rank in range(self.worker_size):
            group_id=(self.graph.partptr[rank]<=node_id_list) & (node_id_list<self.graph.partptr[rank+1])
            part_list.append(torch.masked_select(node_id_list,group_id))
            scatter_list.append(torch.nonzero(group_id))
            #index=torch.nonzero(group_id)
            #scatter_list=torch.stack((scatter_list,index))
        scatter_index=torch.cat(scatter_list)
        return part_list,scatter_index
    
    def get_random_root(self,num_samples):
        print(num_samples)
        return self.graph.get_globalId_by_partitionId(self.rank,torch.randint(high=self.graph.get_part_num(),size=(num_samples,)))

    def sampleGraph(self,sampler,num_samples,**kwargs):
        print('sample',num_samples)
        t1=time.time()
        root_list=self.get_random_root(num_samples)
        print('root_list',root_list)
        t2=time.time()
        #邻居节点list和edge list
        if('num_neighbors' in kwargs):
            num_neighbors=kwargs.get('num_neighbors')
            #neighbors=root_list
            #edge_index=[]
            #print( self.graph.get_node_num(),num_neighbors)
            neighbors, edge_index = sampler.sample_from_nodes(root_list,self.graph.edge_index, self.graph.get_node_num(),num_neighbors)
            #print(neighbors)
        t3=time.time()
        part_feature_list,scatter_list=self.split_node_part(neighbors)
        t4=time.time()
        node_feature=self.get_attr(part_feature_list)
        t5=time.time()
        nids=neighbors
        attr_size=node_feature.size()
        x=torch.zeros(attr_size)
        x=x.scatter(0,scatter_list.repeat(1,attr_size[1]),node_feature)
        print(nids)
        #root_list=nids[:num_samples]
        print(root_list)
        local_y_id=self.graph.get_localId_by_partitionId(self.rank,root_list)
        y=self.graph.select_y(local_y_id)
        print(t2-t1,t3-t2,t4-t3,t5-t4,time.time()-t4)
        return BatchData(num_samples,nids,edge_index,roots=root_list,x=x,y=y)

    def get_attr(self,node_part_list):
        #worker=self.worker#sampler_worker[self.rank]
        #global local_sample_worker
        worker=local_sample_worker[self.rank]
        part_feature=worker.request_attr(self.data_name,node_part_list)
        local_feature=self.get_localattr(node_part_list[self.rank])
        node_feature_list=[]
        for i in range(self.worker_size):
            if(i<self.rank):
                node_feature_list.append(part_feature[i])
            elif i>self.rank:
                node_feature_list.append(part_feature[i-1])
            else:
                node_feature_list.append(local_feature)
        return torch.cat(node_feature_list)
        
    def get_localattr(self,node_list):
        local_id=self.graph.get_localId_by_partitionId(self.rank,node_list)
        return self.graph.select_attr(local_id)

if __name__=="__main__":
    worker=sampler_worker(2,"observe")
    worker.loadGraph('./part/metis_2')
    sampler=NeighborSampler()
    worker.sampler([],sampler,5,num_neighbors=3)